import warnings

from typing import Any

import pandas as pd

import tabulate

from .column_dispatch import describe_columns


class TextFormatter():
    def __init__(self, df: pd.DataFrame, width: int = 100):
        self.df = df
        self.width = width
        self.sep_char = '─'

    def _render_header(self, title: str) -> str:
        header = f'{self.sep_char * 2} {title} '
        header += self.sep_char * (self.width - len(header))
        return header

    def _render_divider(self) -> str:
        return self.sep_char * (self.width // 2) + '\n'

    def _render_dataframe_as_table(
        self, df: pd.DataFrame, **kwargs: Any
    ) -> str:
        # , floatfmt='.3f'
        return tabulate.tabulate(df, headers='keys', **kwargs)

    def render_dataframe_overview(self) -> str:
        txt = ''

        header = self._render_header('Data Summary')
        txt += f'{header}\n'

        txt += self._render_dataframe_as_table(pd.DataFrame({
            'type': ['Number of rows', 'Number of columns'],
            'value': self.df.shape
        }).set_index('type'), floatfmt='.0f')
        txt += '\n'

        txt += self._render_divider()

        dtype_freqs = self.df.dtypes.value_counts().to_frame('Count')
        txt += 'Column type frequency:\n'
        txt += self._render_dataframe_as_table(dtype_freqs)
        txt += '\n\n'

        return txt

    def render_variable_summaries(self) -> str:
        txt = ''

        # generate descriptions
        described_columns = set()
        for type_, df_summary in describe_columns(self.df):
            if df_summary is None:
                continue
            described_columns.update(df_summary['name'])

            header = self._render_header(f'Variable type: {type_}')
            txt += f'{header}\n'

            txt += self._render_dataframe_as_table(df_summary, floatfmt='.3g')
            txt += '\n\n'

        # check if all columns were described
        undescribed_columns = set(self.df.columns) - described_columns
        if len(undescribed_columns) > 0:
            warnings.warn(
                f'The following columns had no matching descriptor: {undescribed_columns}')

        return txt


def skim(df: pd.DataFrame, convert_dtypes=False) -> None:
    if convert_dtypes:
        df = df.convert_dtypes()

    if isinstance(df, pd.core.groupby.generic.DataFrameGroupBy):
        print('wow')

    fmtr = TextFormatter(df)

    print(fmtr.render_dataframe_overview(), end='')
    print(fmtr.render_variable_summaries(), end='')



import warnings
from typing import Any, Dict, Callable, Iterator, Tuple

import numpy as np
import pandas as pd

from .utils import text_histogram, top_counts


def describe_numeric(column: pd.Series) -> Dict[str, Any]:
    return {
        'mean': column.mean(),
        'sd': column.std(),
        **{f'p{q*100:.0f}': column.quantile(q=q)
           for q in [0, .25, .5, .75, 1]},
        'hist': text_histogram(column)
    }


def describe_categorical(column: pd.Series) -> Dict[str, Any]:
    return {
        'n_unique': column.nunique(),
        'top_counts': top_counts(column)
    }


def describe_boolean(column: pd.Series) -> Dict[str, Any]:
    return {
        'mean': column.mean(),
        'top_counts': top_counts(column)
    }


def describe_datetime(column: pd.Series) -> Dict[str, Any]:
    return {
        'n_unique': column.nunique(),
        'mean': column.mean(),
        'min': column.min(),
        'max': column.max()
    }


def describe_column(
    column: pd.Series,
    func: Callable[[pd.Series], Dict[str, Any]]
) -> Dict[str, Any]:
    """Provide statistics useful for all dtypes and call descriptor."""
    # clean column
    if np.inf in column.dropna().values or -np.inf in column.dropna().values:
        msg = f'Column "{column.name}" contains +-Inf, replacing with NA.'
        warnings.warn(RuntimeWarning(msg))

        column = column.replace([np.inf, -np.inf], pd.NA)

    # compute statistics
    return {
        'name': column.name,
        'na_count': column.isna().sum(),  # / column.shape[0]
        **func(column)
    }


def describe_columns(df: pd.DataFrame) -> Iterator[Tuple[str, pd.DataFrame]]:
    DISPATCH_MAP = {
        'boolean': describe_boolean,
        'number': describe_numeric,
        'category': describe_categorical,
        'string': describe_categorical,
        'datetime': describe_datetime,
        'object': describe_categorical
    }

    for type_, func in DISPATCH_MAP.items():
        df_sub = df.select_dtypes(include=type_)
        if df_sub.empty:
            yield type_, None
            continue

        df_summary = pd.DataFrame(
            df_sub.apply(lambda x: describe_column(x, func)).to_list())
        yield type_, df_summary



from typing import List, Union

import numpy as np
import pandas as pd


def text_barchart(values: List[int], safe: bool = True) -> str:
    """Render given values as bars in signle line."""
    # ['▁', '▂', '▃', '▄', '▅', '▆', '▇', '█']
    bar_characters = [chr(i) for i in range(9601, 9609)]

    if safe:
        bar_characters = bar_characters[:-1]

    min_ = min(values)
    max_ = max(values)
    if min_ == max_:
        if min_ == 0:
            # bars should be empty
            max_ = 1
        else:
            # bars should be something non-empty
            min_ -= 1
            max_ += 1
    bins = np.linspace(min_, max_, len(bar_characters) + 1)

    bar_values = pd.cut(
        values, bins=bins, labels=bar_characters, include_lowest=True
    )

    return ''.join(bar_values)


def text_histogram(
    values: np.ndarray,
    bins: Union[int, List[int]] = 10
) -> str:
    """Render histogram of given `values` in single line."""
    hist, _ = np.histogram(values.dropna(), bins=bins)
    return text_barchart(hist)


def shorten_str(text: str, width: int = 30, suffix: str = '[...]') -> str:
    """Custom string shortening (`textwrap.shorten` collapses whitespace)."""
    if len(text) <= width:
        return text
    else:
        return text[:width-len(suffix)] + suffix


def top_counts(column: pd.Series, num: int = 3) -> str:
    """Render values of column with highest counts to string."""
    return ', '.join(
        f'{shorten_str(str(k))}: {v}'
        for k, v in column.value_counts().head(num).items()
    )


if __name__ == '__main__':
    txt = text_histogram(pd.Series(np.r_[
        np.random.normal(10, 2, size=500),
        np.random.normal(50, 7, size=500)
    ]))
    print(txt)



import pandas as pd

import seaborn as sns
import matplotlib.pyplot as plt


def annotate_correlation(*args, method, **kwargs):
    """Plot correlation.

    Adapted from https://github.com/mwaskom/seaborn/issues/1444
    """
    # TODO: handle hue

    # compute correlation
    corr_r = args[0].corr(args[1], method)
    corr_text = f'{corr_r:2.2f}'.replace('0.', '.')

    # visualize correlation
    ax = plt.gca()
    ax.set_axis_off()

    marker_size = abs(corr_r) * 10000
    ax.scatter(
        .5, .5, marker_size, corr_r, alpha=0.6,
        cmap='vlag_r', vmin=-1, vmax=1,  # bwr_r
        transform=ax.transAxes
    )

    ax.annotate(
        corr_text,
        [.5, .5], xycoords='axes fraction',
        ha='center', va='center', fontsize=20
    )

    return ax


def corrplot(
    df: pd.DataFrame,
    corr_method: str = 'spearman', **kwargs
) -> sns.PairGrid:
    """An improved version of `sns.pairplot`."""
    g = sns.PairGrid(df, **kwargs)

    g.map_upper(annotate_correlation, method=corr_method)
    g.map_diag(sns.distplot, kde=False)
    g.map_lower(sns.scatterplot, rasterized=True)  # regplot

    return g



import pandas as pd

import click

from .text_formatter import skim


@click.command(help='Quickly create summary statistics for a given dataframe.')
@click.option('-d', '--delimiter', default=',', help='Delimiter of file.')
@click.option(
    '-i', '--interactive', is_flag=True,
    help='Open prompt with dataframe as `df` after displaying summary.'
)
@click.option(
    '--no-dtype-conversion', is_flag=True, default=False,
    help='Skip automatic dtype conversion.'
)
@click.argument(
    'fname', type=click.Path(exists=True, dir_okay=False),
    metavar='<file>'
)
def main(
    delimiter: str, interactive: bool, no_dtype_conversion: bool,
    fname: str
) -> None:
    # read dataframe
    df = pd.read_csv(fname, sep=delimiter, low_memory=False)

    # na_characters = []
    # if len(na_characters) > 0:
    #     df.replace(na_characters, pd.NA, inplace=True)

    if not no_dtype_conversion:
        # do type conversion here for availability in interactive mode
        df = df.convert_dtypes()

    # skim dataframe
    skim(df)

    # interactive mode
    if interactive:
        # reduce loading time when interactive mode is not enabled
        import IPython

        IPython.embed()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyskim']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<9.0.0',
 'ipython>=7.19.0,<8.0.0',
 'numpy>=1.19.4,<2.0.0',
 'pandas>=1.1.5,<2.0.0',
 'scipy>=1.5.4,<2.0.0',
 'tabulate>=0.8.7,<0.9.0']

entry_points = \
{'console_scripts': ['pyskim = pyskim:main']}

setup_kwargs = {
    'name': 'pyskim',
    'version': '0.1.3',
    'description': 'Quickly create summary statistics for a given dataframe.',
    'long_description': "# pyskim\n\n[![PyPI](https://img.shields.io/pypi/v/pyskim.svg?style=flat)](https://pypi.python.org/pypi/pyskim)\n\nQuickly create summary statistics for a given dataframe.\n\nThis package aspires to be as awesome as [skimr](https://github.com/ropensci/skimr).\n\n\n\nhttps://github.com/ropensci/skimr\nhttps://github.com/ropensci/skimr/blob/master/R/get_skimmers.R\nhttps://github.com/b-rodrigues/brotools/blob/master/R/describe.R\n\nhttps://pandas.pydata.org/pandas-docs/stable/user_guide/basics.html#dtypes\nhttps://numpy.org/doc/stable/reference/arrays.scalars.html\n\nhttps://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Int64Dtype.html\n\n\n\n## Installation\n\n```bash\n$ pip install pyskim\n```\n\n## Usage\n\n### Commandline tool\n\n\n`pyskim` can be used from the commandline:\n\n```bash\n$ pyskim iris.csv\n── Data Summary ────────────────────────────────────────────────────────────────────────────────────\ntype                 value\n-----------------  -------\nNumber of rows         150\nNumber of columns        5\n──────────────────────────────────────────────────\nColumn type frequency:\n           Count\n-------  -------\nfloat64        4\nstring         1\n\n── Variable type: number ───────────────────────────────────────────────────────────────────────────\n    name            na_count    mean     sd    p0    p25    p50    p75    p100  hist\n--  ------------  ----------  ------  -----  ----  -----  -----  -----  ------  ----------\n 0  sepal_length           0    5.84  0.828   4.3    5.1   5.8     6.4     7.9  ▂▆▃▇▄▇▅▁▁▁\n 1  sepal_width            0    3.06  0.436   2      2.8   3       3.3     4.4  ▁▁▄▅▇▆▂▂▁▁\n 2  petal_length           0    3.76  1.77    1      1.6   4.35    5.1     6.9  ▇▃▁▁▂▅▆▄▃▁\n 3  petal_width            0    1.2   0.762   0.1    0.3   1.3     1.8     2.5  ▇▂▁▂▂▆▁▄▂▃\n\n── Variable type: string ───────────────────────────────────────────────────────────────────────────\n    name               na_count    n_unique  top_counts\n--  ---------------  ----------  ----------  -----------------------------------------\n 0          species           0           3  versicolor: 50, setosa: 50, virginica: 50\n```\n\nFull overview:\n\n```bash\n$ pyskim --help\nUsage: pyskim [OPTIONS] <file>\n\n  Quickly create summary statistics for a given dataframe.\n\nOptions:\n  -d, --delimiter TEXT   Delimiter of file.\n  -i, --interactive      Open prompt with dataframe as `df` after displaying\n                         summary.\n\n  --no-dtype-conversion  Skip automatic dtype conversion.\n  --help                 Show this message and exit.\n```\n\n### Python API\n\nAlternatively, it is possible to use it in code:\n\n```python\n>>> from pyskim import skim\n>>> from seaborn import load_dataset\n\n>>> iris = load_dataset('iris')\n>>> skim(iris)\n# ── Data Summary ────────────────────────────────────────────────────────────────────────────────────\n# type                 value\n# -----------------  -------\n# Number of rows         150\n# Number of columns        5\n# ──────────────────────────────────────────────────\n# Column type frequency:\n#            Count\n# -------  -------\n# float64        4\n# string         1\n#\n# ── Variable type: number ───────────────────────────────────────────────────────────────────────────\n#     name            na_count    mean     sd    p0    p25    p50    p75    p100  hist\n# --  ------------  ----------  ------  -----  ----  -----  -----  -----  ------  ----------\n#  0  sepal_length           0    5.84  0.828   4.3    5.1   5.8     6.4     7.9  ▂▆▃▇▄▇▅▁▁▁\n#  1  sepal_width            0    3.06  0.436   2      2.8   3       3.3     4.4  ▁▁▄▅▇▆▂▂▁▁\n#  2  petal_length           0    3.76  1.77    1      1.6   4.35    5.1     6.9  ▇▃▁▁▂▅▆▄▃▁\n#  3  petal_width            0    1.2   0.762   0.1    0.3   1.3     1.8     2.5  ▇▂▁▂▂▆▁▄▂▃\n#\n# ── Variable type: string ───────────────────────────────────────────────────────────────────────────\n#     name               na_count    n_unique  top_counts\n# --  ---------------  ----------  ----------  -----------------------------------------\n#  0          species           0           3  versicolor: 50, setosa: 50, virginica: 50\n```\n",
    'author': 'kpj',
    'author_email': 'kim.philipp.jablonski@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kpj/pyskim',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
